import datetime
import logging
import re
from time import time
from typing import Literal

from drafteleu.bump import bump_version
from drafteleu.changelog import compile_changelog, get_prs
from drafteleu.releases import find_latest_release, create_or_update_draft_release

logger = logging.getLogger(__name__)


def preload_prs(repo_slug: str, configurations):
    start = time()

    all_labels = set()
    earliest_release_date = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")
    for config in configurations:
        include_labels = config.get("include_labels", [])
        all_labels.update(include_labels)
        tag_regex = config.get("tag_regex", "v(?P<version>.+)")
        stage = config.get("stage", "both")

        for current_stage in ("pre", "final") if stage == "both" else (stage,):
            include_prerelease = current_stage == "pre"
            latest_release = find_latest_release(
                tag_regex,
                repo=repo_slug,
                include_prereleases=include_prerelease,
            )
            if latest_release is None:
                logger.info("No previous release found, assuming initial release")
                earliest_release_date = "1970-01-01T00:00:00Z"
                # No point in looking for earlier releases
                break
            else:
                release_date = latest_release["publishedAt"]
            if release_date < earliest_release_date:
                earliest_release_date = release_date

    # Now that we have the earliest release date and all labels, fetch PRs once
    prs = get_prs(
        merged_since=earliest_release_date,
        repo=repo_slug,
        labels=list(all_labels),
    )
    logger.info(
        f"Preloaded {len(prs)} PRs since {earliest_release_date} in {time() - start:.2f}s"
    )
    return prs


def pr_has_any_label(pr, labels: set) -> bool:
    return not labels or labels.intersection(pr["labels"])


def drafteleu(
    repo_slug: str,
    *,
    stage: Literal["pre", "final", "both"] = "both",
    dry_run: bool = False,
    configurations: list[dict],
):
    all_prs = preload_prs(repo_slug, configurations)

    for config in configurations:
        version_template = config.get("version_template", "{major}.{minor}.{patch}")
        tag_template = config.get("tag_template", "v{next_version}")
        title_template = config.get("title_template", "Release {next_version}")
        tag_regex = config.get("tag_regex", "v(?P<version>.+)")
        rc_template = config.get("rc_template", "-{pre_label}{pre_number}")
        bump_part = config.get("bump_part", "minor")
        include_labels = set(config.get("include_labels", []))
        for current_stage in ("pre", "final") if stage == "both" else (stage,):
            logger.info(f"Processing stage: {current_stage}")
            include_prerelease = current_stage == "pre"

            latest_release = find_latest_release(
                tag_regex, repo=repo_slug, include_prereleases=include_prerelease
            )
            if latest_release is None:
                logger.info("No previous release found, assuming initial release")
                initial_version = version_template.format(major=0, minor=0, patch=0)
                latest_release = {
                    "tagName": tag_template.format(next_version=initial_version),
                    "publishedAt": "1970-01-01T00:00:00Z",
                }
            else:
                logger.info(
                    f"Latest release: {latest_release['tagName']} at {latest_release['publishedAt']}"
                )

            prs = [
                pr
                for pr in all_prs
                if pr_has_any_label(pr, include_labels)
                and pr["closedAt"] > latest_release["publishedAt"]
            ]
            logger.info(
                f"Found {len(prs)} PRs since last release with labels {include_labels}"
            )

            changelog = compile_changelog(prs)
            logger.debug("Changelog:\n" + changelog)

            current_version = re.match(tag_regex, latest_release["tagName"]).group(
                "version"
            )
            logger.info(f"Current version: {current_version}")
            next_version = bump_version(
                current_version,
                "dev" if include_prerelease else bump_part,
                version_template=version_template,
                rc_template=rc_template,
                rc_bump_part=bump_part,
            )
            next_tag = tag_template.format(next_version=next_version)
            next_title = title_template.format(
                next_version=next_version,
                next_tag=next_tag,
                current_version=current_version,
                current_tag=latest_release["tagName"],
                current_date=latest_release["publishedAt"],
                stage=current_stage,
            )
            logger.info(f"Next version: {next_version}")
            logger.info(f"Next tag: {next_tag}")
            logger.info(f"Next title: {next_title}")

            create_or_update_draft_release(
                next_tag,
                next_title,
                changelog,
                repo=repo_slug,
                dry_run=dry_run,
                prerelease=include_prerelease,
            )



import json
import logging
import re
import subprocess
from functools import cache
from packaging.version import Version

logger = logging.getLogger(__name__)


@cache
def fetch_releases(repo: str, *, include_prereleases: bool = True) -> list[dict]:
    result = subprocess.run(
        [
            "gh",
            "release",
            "list",
            "--limit",
            "1000",
            "--json",
            "tagName,publishedAt",
            "--exclude-drafts",
            *(["--repo", repo] if repo else []),
            *(["--exclude-pre-releases"] if not include_prereleases else []),
        ],
        capture_output=True,
        text=True,
        check=False,
    )
    if result.returncode != 0:
        raise RuntimeError(f"gh command failed: {result.stderr}")
    releases = json.loads(result.stdout)
    return releases


def find_latest_release(
    tag_regex: str, include_prereleases=True, repo=None
) -> dict | None:
    releases = fetch_releases(repo, include_prereleases=include_prereleases)

    if tag_regex:
        relevant_releases = [r for r in releases if re.match(tag_regex, r["tagName"])]
    else:
        relevant_releases = releases

    relevant_releases.sort(
        key=lambda r: Version(re.match(tag_regex, r["tagName"]).group("version")),
        reverse=True,
    )

    latest_release = relevant_releases[0] if relevant_releases else None

    return latest_release


def create_or_update_draft_release(
    tag_name: str,
    title: str,
    body: str,
    repo: str = None,
    dry_run: bool = False,
    prerelease: bool = False,
) -> None:
    result = subprocess.run(
        [
            "gh",
            "release",
            "view",
            tag_name,
            *(["--repo", repo] if repo else []),
            "--json",
            "tagName,name,body",
        ],
        capture_output=True,
        text=True,
        check=False,
    )
    exists = result.returncode == 0

    if exists:
        existing_release = json.loads(result.stdout)
        needs_update = not all(
            (
                tag_name == existing_release["tagName"],
                title.strip() == existing_release["name"].strip(),
                body.strip() == existing_release["body"].strip(),
            )
        )
        if not needs_update:
            logger.info(f"No changes for release {tag_name}, skipping update.")
            return
        elif dry_run:
            logger.info(
                f"Would update draft release {tag_name} in repo {repo or 'default'} with title '{title}' and body:\n{body}\n"
            )
            return

        logger.info(f"Updating existing release {tag_name}.")

        cmd = [
            "gh",
            "release",
            "edit",
            tag_name,
            *(["--repo", repo] if repo else []),
            *("--title", title),
            *("--notes", body),
            *(["--prerelease"] if prerelease else []),
        ]
    else:
        if dry_run:
            logger.info(
                f"Would create draft release {tag_name} in repo {repo or 'default'} with title '{title}' and body:\n{body}\n"
            )
            return

        logger.info(f"Creating new draft release {tag_name}.")

        cmd = [
            "gh",
            "release",
            "create",
            tag_name,
            "--draft",
            *(["--repo", repo] if repo else []),
            *("--title", title),
            *("--notes", body),
            *(["--prerelease"] if prerelease else []),
        ]

    result = subprocess.run(cmd, capture_output=True, text=True, check=False)
    if result.returncode != 0:
        raise RuntimeError(f"gh command failed: {result.stderr}")



import argparse
import re

from typing import Literal

from packaging.version import parse


def make_version(
    major: int,
    minor: int | None,
    patch: int | None,
    pre: tuple[str, int] | None,
    version_template: str = "{major}.{minor}.{patch}",
    rc_template: str = "-{pre_label}{pre_number}",
) -> str:
    version = version_template.format(major=major, minor=minor, patch=patch)
    if pre is not None:
        version += rc_template.format(pre_label=pre[0], pre_number=pre[1])

    return version


def parse_full_version(version_string: str) -> tuple[str, str, str]:
    parts = re.split(r"(/v?)", version_string, maxsplit=1)
    if len(parts) == 1:
        prefix, separator, version_string = ("", "", parts[0])
        if version_string.startswith("v"):
            prefix = "v"
            version_string = version_string[1:]
    else:
        prefix, separator, version_string = parts
    parse(version_string)  # Validate
    return prefix, separator, version_string


def bump_version(
    version_string: str,
    part: Literal["major", "minor", "patch", "dev"],
    version_template: str = "{major}.{minor}.{patch}",
    rc_template="-{pre_label}{pre_number}",
    rc_bump_part: Literal["major", "minor", "patch"] = "minor",
) -> str:
    leading_v = "v" if version_string.startswith("v") else ""
    version = parse(version_string)

    major, minor, patch, *_ = version.release + (0,) * 2
    pre = version.pre
    if part == "major":
        major, minor, patch = major + 1, 0, 0
    elif part == "minor":
        major, minor, patch = major, minor + 1, 0
    elif part == "patch":
        major, minor, patch = major, minor or 0, patch + 1
    elif part == "dev":
        if version.is_prerelease:
            if pre is None:
                # Could be .devN
                pre_label, pre_num = "dev", version.dev
            else:
                pre_label, pre_num = pre
            pre_num += 1
        else:
            pre_label, pre_num = "rc", 0
            # Which part to bump?
            if rc_bump_part == "major":
                major, minor, patch = major + 1, 0, 0
            elif rc_bump_part == "minor":
                major, minor, patch = major, minor + 1, 0
            elif rc_bump_part == "patch":
                major, minor, patch = major, minor, patch + 1
            else:
                raise ValueError(f"Unknown rc_bump_part: {rc_bump_part}")

        pre = (pre_label, pre_num)
    else:
        raise ValueError(f"Unknown part to bump: {part}")

    new_version = make_version(major, minor, patch, pre, version_template, rc_template)

    return f"{leading_v}{new_version}"


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Bump version script")
    parser.add_argument(
        "version",
        type=str,
        help="The new version to set (e.g., '1.2.3'). Supports prefixes like 'app/v1.2.3'",
    )
    parser.add_argument(
        "part",
        type=str,
        choices=["major", "minor", "patch", "dev"],
        help="Which part of the version to bump",
    )

    args = parser.parse_args()
    print(bump_version(args.version, args.part))



import argparse
import logging
from pathlib import Path

import yaml

from .drafteleu import drafteleu


def main():
    parser = argparse.ArgumentParser(description="Drafteleu CLI")
    parser.add_argument(
        "repo_slug", type=str, help="GitHub repository slug (e.g., user/repo)"
    )
    parser.add_argument(
        "--stage",
        choices=["pre", "final", "both"],
        default="both",
        help="Process only prereleases, only final releases, or both (default: both)",
    )
    parser.add_argument(
        "--verbosity",
        "-v",
        type=int,
        choices=[0, 1, 2],
        default=1,
        help="Set verbosity level: 0 = none, 1 = info, 2 = debug (default: 1)",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Run the script without making any changes (for testing)",
    )
    parser.add_argument(
        "--config",
        type=Path,
        default="drafteleu.yaml",
        help="Path to configuration file (optional)",
    )

    args = parser.parse_args()

    if not args.config.exists():
        raise FileNotFoundError(f"Configuration file not found: {args.config}")

    if args.verbosity == 0:
        logging.basicConfig(level=logging.WARNING)
    elif args.verbosity == 1:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.DEBUG)

    configurations = list(
        yaml.safe_load_all(args.config.read_text()) if args.config else []
    )

    drafteleu(
        args.repo_slug,
        stage=args.stage,
        dry_run=args.dry_run,
        configurations=configurations,
    )


if __name__ == "__main__":
    main()



import argparse
import json
import operator
import subprocess
import sys
from datetime import datetime
from pathlib import Path
from typing import Iterable, Callable, Hashable

from drafteleu.releases import find_latest_release


PATH_TO_GRAPHQL_QUERY = Path(__file__).parent.joinpath("queries")


def transform_pr(pr):
    pr["labels"] = [label["name"] for label in pr.get("labels", {}).get("nodes", [])]
    return pr


def unique_by_key[T](
    iterable: Iterable[T], key: Callable[[T], Hashable]
) -> Iterable[T]:
    """
    Remove duplicates from an iterable based on a key function.
    """
    seen = {}
    for item in iterable:
        if (k := key(item)) not in seen:
            seen[k] = item
    return seen.values()


def graphql_query(query_name: str, **kwargs):
    args = [
        "gh",
        "api",
        "graphql",
        "-F",
        f"query=@{PATH_TO_GRAPHQL_QUERY.joinpath(f'{query_name}.graphql').as_posix()}",
    ]
    for k, v in kwargs.items():
        if isinstance(v, list):
            for item in v:
                args.extend(["-F", f"{k}[]={item}"])
        elif v is not None:
            args.extend(["-F", f"{k}={v}"])
    response = subprocess.run(
        args,
        capture_output=True,
        text=True,
        check=False,
    )
    if response.returncode != 0:
        raise RuntimeError(f"gh command failed: {response.stderr}")

    return json.loads(response.stdout)


def get_prs(labels: list[str], *, merged_since: str | None = None, repo: str):
    owner, repo_name = repo.split("/")

    min_closed_at = datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")
    after = None
    all_prs = []

    while merged_since and merged_since < min_closed_at:
        page = graphql_query(
            "get_prs",
            repository=repo_name,
            owner=owner,
            labels=labels,
            after=after,
        )
        prs = [
            transform_pr(item["node"])
            for item in page["data"]["repository"]["pullRequests"]["edges"]
            if item["node"].get("closedAt")
        ]

        min_closed_at = min((pr["closedAt"] for pr in prs), default=min_closed_at)

        all_prs.extend(prs)
        page_info = page["data"]["repository"]["pullRequests"]["pageInfo"]
        if page_info["hasNextPage"]:
            after = page_info["endCursor"]
        else:
            break

    if merged_since:
        all_prs = [pr for pr in all_prs if pr["closedAt"] > merged_since]

    return unique_by_key(all_prs, key=operator.itemgetter("number"))


def has_label(pr, label):
    return label in pr["labels"]


def format_author(author):
    return author["login"].removeprefix("app/")


def compile_changelog(prs: list[dict]) -> str:
    changelog = "## What’s Changed\n\n"
    for pr in filter(lambda pr: not has_label(pr, "dependencies"), prs):
        changelog += (
            f"* {pr['title']} (#{pr['number']}) @{format_author(pr['author'])}\n"
        )

    dependencies = [pr for pr in prs if has_label(pr, "dependencies")]
    if dependencies:
        changelog += "\n## :arrow_up: Dependencies\n\n"
        for pr in dependencies:
            changelog += (
                f"* {pr['title']} (#{pr['number']}) @{format_author(pr['author'])}\n"
            )
    return changelog


def main(app: str, include_prereleases: bool, repo: str | None) -> None:
    latest_release = find_latest_release(
        tag_regex=rf"^{app}/", include_prereleases=include_prereleases, repo=repo
    )
    sys.stderr.write(
        f"Latest release: {latest_release['tagName']} at {latest_release['publishedAt']}\n"
    )
    prs = get_prs([app], merged_since=latest_release["publishedAt"])
    sys.stderr.write(f"Found {len(prs)} PRs since last release\n")
    changelog = compile_changelog(prs)
    sys.stdout.write(changelog + "\n")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        "Generate changelog for an app or package since its last release",
    )
    parser.add_argument(
        "app",
        type=str,
        help="The app or package name to generate the changelog for",
    )
    parser.add_argument(
        "--include-prereleases",
        "--pre",
        action="store_true",
        help="Whether to include pre-releases when determining the latest release",
    )
    parser.add_argument(
        "--repo",
        "-r",
        type=str,
        help='GitHub repository in the format "owner/repo". Defaults to the current repository.',
        default=None,
    )

    args = parser.parse_args()

    main(args.app, args.include_prereleases, args.repo)
